using SafeTestsets

@safetestset "Memory Tests" begin

    @safetestset "baselevel" begin
        using ACTRModels, Test
        import ACTRModels: baselevel, baselevel!
        include("memory.jl")

        lags = [1]
        d = .5
        @test baselevel(d, lags) == 0

        lags = [1]
        d = .1
        @test baselevel(d, lags) == 0

        lags = [2,10]
        d = .5
        @test baselevel(d, lags) ≈ 0.023066 atol=1e-5

        lags = [.5,10]
        d = .4
        @test baselevel(d, lags) ≈ 0.540936 atol=1e-5
    end

    @safetestset "baselevel!" begin
        using ACTRModels, Test
        import ACTRModels: baselevel, baselevel!
        include("memory.jl")

        actr,memory,chunks = initializeACTR(;bll=true)
        c = get_chunk(actr; animal=:dog,name=:Sigma)[1]
        update_lags!(c, 2.0)
        baselevel!(c, memory)
        @test c.act_bll ≈ -0.346573 atol=1e-5
        c.act_bll = 0.0
        update_lags!(c, 5.0)
        baselevel!(c, memory)
        @test c.act_bll ≈ -0.80471 atol=1e-5
        c.act_bll = 0.0
        update_recent!(c, 10.0)
        update_chunk!(c, 12)
        update_lags!(c, 14)
        baselevel!(c, memory)
        @test c.act_bll ≈ 0.09076 atol=1e-5
        c.act_bll = 0.0
        memory.parms.d = .8
        baselevel!(c, memory)
        @test c.act_bll ≈ -0.220564 atol=1e-5

        actr,memory,chunks = initializeACTR(; bll=true)
        c = get_chunk(actr; animal=:dog, name=:Sigma)[1]
        c.k = 2
        update_chunk!(c, 1)
        update_chunk!(c, 2)
        update_chunk!(c, 4)
        update_lags!(c, 10)
        @test c.N == 4
        @test c.L == 10
        @test all(c.lags .== [6,8])
        @test all(c.recent .== [4,2])
    end

    @safetestset "Chunk" begin
        using ACTRModels, Test
        import ACTRModels: baselevel, baselevel!
        include("memory.jl")

        chunk = Chunk(;N=20, animal=:dog, name=:Sigma)
        @test chunk.slots.animal == :dog
        @test chunk.slots.name == :Sigma
        @test chunk.N == 20
    end

    @safetestset "add Chunk" begin
        using ACTRModels, Test
        import ACTRModels: baselevel, baselevel!
        include("memory.jl")

        actr,memory,chunks = initializeACTR()
        c = get_chunk(actr; animal=:human, name=:Wilford, lastName=:Brimley)
        @test isempty(c)
        add_chunk!(actr, 10.0; animal=:human, name=:Wilford, lastName=:Brimley)
        c = get_chunk(actr; animal=:human, name=:Wilford, lastName=:Brimley)
        @test !isempty(c)
        @test c[1].N == 1
        add_chunk!(actr, 10.0; animal=:human, name=:Wilford, lastName=:Brimley)
        @test c[1].N == 2
    end

    @safetestset "retrieval_prob" begin
        using ACTRModels, Test
        import ACTRModels: baselevel, baselevel!
        include("memory.jl")
        
        actr,memory,chunks = initializeACTR()
        @test retrieval_prob(actr, chunks[2]; animal=:cat, name=:Butters) == (.5,.5)
        @test retrieval_prob(actr, chunks[1]; isa=:mammal) == (0,1)

        actr,memory,chunks = initializeACTR(;τ=.5)
        c = get_chunk(actr; animal=:rat)
        @test retrieval_prob(actr, c; animal=:rat)[1] ≈ 0.38098 atol=1e-5

        actr,memory,chunks = initializeACTR(;τ=.5, mmp=true, δ=1.0)
        c = get_chunk(actr; animal=:rat, name=:Joy)
        p1,_ = retrieval_prob(actr, c)
        @test p1 ≈ 0.137939 atol=1e-5

        p2,_ = retrieval_prob(actr, c; animal=:rat)
        @test p2 ≈ 0.183859 atol=1e-5

        p3,_ = retrieval_prob(actr, c; animal=:rat, name=:Joy)
        @test p3 ≈ 0.229243 atol=1e-5

        p3,_ = retrieval_prob(actr, c; isa=:rock, animal=:rat, name=:Joy)
        @test p3 == 0

        actr,memory,chunks = initializeACTR(;τ=.5, bll=true)
        c = get_chunk(actr; animal=:rat, name=:Joy)
        update_lags!(actr, 1.0)
        p1,_ = retrieval_prob(actr, c, 1.0)
        update_lags!(actr, 3.0)
        p2,_ = retrieval_prob(actr, c, 3.0)
        memory.parms.d = .8
        p3,_ = retrieval_prob(actr, c, 3.0)
        @test p1 > p2 > p3

        actr,memory,chunks = initializeACTR(;d=.5, bll=true)
        c = get_chunk(actr; animal=:rat, name=:Joy)
        update_recent!.(c, 5.0)
        update_lags!(actr, 7)
        p1,_ = retrieval_prob(actr, c, 7.0)
        c[1].N = 2
        p2,_=retrieval_prob(actr, c, 7.0)
        c[1].N = 5
        p3,_ = retrieval_prob(actr, c, 7.0)
        @test p1 < p2 < p3

        actr,memory,chunks = initializeACTR(;sa=true, γ=1.0)
        c = get_chunk(actr; animal=:rat, name=:Joy)
        actr.imaginal.chunk = Chunk(;animal=:rat)
        p1,_ = retrieval_prob(actr, c)
        actr.imaginal.chunk = Chunk(;name=:Joy)
        p2,_ = retrieval_prob(actr, c)
        @test p2 > p1
    end

    @safetestset "modify!" begin
        using ACTRModels, Test
        chunks = Chunk[Chunk(;isa=:bafoon, animal=:dog,name=:Sigma, retrieved=[false]),
        Chunk(;isa=:mammal,animal=:cat,name=:Butters, retrieved=[false])]
        memory = Declarative(;memory=chunks)
        actr = ACTR(;declarative=memory)
        modify!(chunks[2].slots, retrieved=true)
        @test chunks[2].slots.retrieved[1] == true 
    end

    @safetestset "filter" begin
        using ACTRModels, Test
        chunks = Chunk[Chunk(;isa=:bafoon, animal=:dog,name=:Sigma, retrieved=[false]),
        Chunk(;isa=:mammal,animal=:cat,name=:Butters, retrieved=[false])]
        memory = Declarative(;memory=chunks, mmp=true)
        actr = ACTR(;declarative=memory)
        request = retrieval_request(actr; isa=:mammal)
        @test request[1].slots.name == :Butters
    end

    @safetestset "count_values" begin
        using ACTRModels, Test
        import ACTRModels: count_values
        chunk = Chunk(a=:a, b=:a, c=:v)
        @test count_values(chunk, :a) == 2
        @test count_values(chunk, :v) == 1
        @test count_values(chunk, :c) == 0
    end

    @safetestset "update_recent!" begin
        using ACTRModels, Test
        chunk = Chunk(a=:a, b=:a, c=:v)
        update_recent!(chunk, .10)
        @test chunk.recent[1] == .1
        update_recent!(chunk, .20)
        @test chunk.recent[1] == .2
        @test length(chunk.recent) == 1
        chunk.k = 2
        update_recent!(chunk, .10)
        update_recent!(chunk, .20)
        update_recent!(chunk, .30)
        @test chunk.recent[1] == .3
        @test chunk.recent[2] == .2
        @test length(chunk.recent) == 2
    end

    @safetestset "update_lags!" begin
        using ACTRModels, Test
        chunk = Chunk(a=:a, b=:a, c=:v, k=2)
        update_recent!(chunk, .10)
        update_recent!(chunk, .20)
        update_lags!(chunk, 1.0)
        @test chunk.lags[1] == 1 - .2
        @test chunk.lags[2] == 1 - .1
    end

    @safetestset "update_chunk!" begin
        using ACTRModels, Test
        chunk = Chunk(a=:a, b=:a, c=:v)
        update_chunk!(chunk, .10)
        @test chunk.recent[1] == .1
        @test chunk.N == 2
        update_chunk!(chunk, .20)
        @test chunk.recent[1] == .2
        @test chunk.N == 3
        @test length(chunk.recent) == 1
    end

    @safetestset "spreading activation" begin
        using ACTRModels, Test
        chunks = [Chunk(a=:a, b=:b, c=:c), Chunk(a=:a, b=:b, c=:a)]
        memory = Declarative(memory=chunks, γ=1.6)
        imaginal = Imaginal(chunk=Chunk(a=:a, b=:b))
        actr = ACTR(declarative=memory, imaginal=imaginal)
        compute_activation!(actr)
        @test chunks[1].act == 0.0
        @test chunks[2].act == 0.0
        memory.parms.sa = true
        compute_activation!(actr)
        @test chunks[1].act ≈ 0.3575467 atol = 1e-5
        @test chunks[2].act ≈ 0.7041203 atol = 1e-5
    end

    @safetestset "partial matching" begin
        using ACTRModels, Test
        chunk = Chunk(a=:a, b=:b)
        memory = Declarative(memory=[chunk], mmp=true, δ=1.0)
        actr = ACTR(declarative=memory)
        compute_activation!(actr)
        @test chunk.act == 0.0
        compute_activation!(actr; a=:a)
        @test chunk.act == 0.0
        compute_activation!(actr; a=:a, b=:b)
        @test chunk.act == 0.0
        compute_activation!(actr; a=:b, b=:b)
        @test chunk.act == -1.0
        compute_activation!(actr; a=:a, b=:a)
        @test chunk.act == -1.0
        compute_activation!(actr; a=:c, b=:c)
        @test chunk.act == -2.0
    end 

    @safetestset "get_chunk" begin
        using ACTRModels, Test
        chunks = [Chunk(a=:a, b=:c),Chunk(a=:a, b=:b)]
        memory = Declarative(memory=chunks)
        actr = ACTR(declarative=memory)
        result = get_chunk(actr; a=:b)
        @test isempty(result)

        result = get_chunk(actr; a=:a)
        @test !isempty(result)
        @test length(result) == 2
        @test result[1].slots.a == :a
        @test result[1].slots.b == :c
        @test result[2].slots.a == :a
        @test result[2].slots.b == :b

       result = get_chunk(actr; a=:a, b=:b)
       @test !isempty(result)
       @test length(result) == 1
       @test result[1].slots.a == :a
       @test result[1].slots.b == :b
    end 

    @safetestset "first_chunk" begin
        using ACTRModels, Test
        chunks = [Chunk(a=:a, b=:c),Chunk(a=:a, b=:b)]
        memory = Declarative(memory=chunks)
        actr = ACTR(declarative=memory)
        result = get_chunk(actr; a=:b)
        @test isempty(result)

        result = first_chunk(actr; a=:a)
        @test !isempty(result)
        @test length(result) == 1
        @test result[1].slots.a == :a
        @test result[1].slots.b == :c
    end 

    @safetestset "reset_activation!" begin
        using ACTRModels, Test
        import ACTRModels: reset_activation!
        chunks = [Chunk(a=:a, b=:b, c=:c), Chunk(a=:a, b=:b, c=:a)]
        chunk = chunks[1]
        memory = Declarative(memory=chunks, mmp=true, noise=true, bll=true, sa=true, γ=1.6,
            δ=1.0)
        imaginal = Imaginal(chunk=Chunk(a=:a, b=:b))
        actr = ACTR(declarative=memory, imaginal=imaginal)
        compute_activation!(actr, 3.0; a=:b)
        @test chunk.act_bll != 0
        @test chunk.act_pm != 0
        @test chunk.act_sa != 0
        @test chunk.act_noise != 0
        reset_activation!(chunk)
        @test chunk.act_bll == 0
        @test chunk.act_pm == 0
        @test chunk.act_sa == 0
        @test chunk.act_noise == 0 
    end

    @safetestset "total_activation!" begin
        using ACTRModels, Test
        import ACTRModels: total_activation!
        chunk = Chunk()
        chunk.act_blc = .5
        chunk.act_bll = 1.0
        chunk.act_pm = 1.0
        chunk.act_sa = 1.0
        chunk.act_noise = 1.0
        total_activation!(chunk)
        @test chunk.act ≈ 2.5 atol = 1e-5
    end
end